import boto3
import pydash
from copy import deepcopy

from infra_buddy.utility import print_utility, waitfor

from infra_buddy.aws.cloudformation import CloudFormationBuddy


class ECSBuddy(object):
    def __init__(self, deploy_ctx):
        # type: (DeployContext) -> None
        super(ECSBuddy, self).__init__()
        self.deploy_ctx = deploy_ctx
        self.client = boto3.client('ecs', region_name=self.deploy_ctx.region)
        cf = CloudFormationBuddy(deploy_ctx)
        ecs_cluster_export_key = "{}-ECSCluster".format(self.deploy_ctx.cluster_stack_name)
        self.cluster = self._wait_for_export(cf=cf,fully_qualified_param_name=ecs_cluster_export_key)
        ecs_service_export_key = "{}-ECSService".format(self.deploy_ctx.stack_name)
        self.ecs_service = self._wait_for_export(cf=cf,fully_qualified_param_name=ecs_service_export_key)
        ecs_task_family_export_key = "{}-ECSTaskFamily".format(self.deploy_ctx.stack_name)
        self.ecs_task_family = self._wait_for_export(cf=cf,fully_qualified_param_name=ecs_task_family_export_key)
        self.task_definition_description = None
        self.new_image = None

    def _wait_for_export(self,cf,fully_qualified_param_name):
        # we are seeing an issue where immediately after stack create the export values are not
        # immediately available
        return waitfor.waitfor(function_pointer=cf.get_export_value,
                               expected_result=None,
                               interval_seconds=2,
                               max_attempts=5,
                               negate=True,
                               args={"fully_qualified_param_name":fully_qualified_param_name},
                               exception=False)

    def set_container_image(self,location,tag):
        self.new_image = "{location}:{tag}".format(location=location,tag=tag)

    def requires_update(self):
        if not self.new_image:
            print_utility.warn("Checking for ECS update without registering new image ")
            return False
        self._describe_task_definition()
        existing = pydash.get(self.task_definition_description, "containerDefinitions[0].image")
        print_utility.info("ECS task existing image - {}".format(existing))
        print_utility.info("ECS task desired image - {}".format(self.new_image))
        return existing != self.new_image

    def perform_update(self):
        self._describe_task_definition(refresh=True)
        new_task_def = {
            'family':self.task_definition_description['family'],
            'containerDefinitions':self.task_definition_description['containerDefinitions'],
            'volumes':self.task_definition_description['volumes']
        }
        new_task_def['containerDefinitions'][0]['image'] = self.new_image
        if 'TASK_MEMORY' in self.deploy_ctx and self.deploy_ctx['TASK_MEMORY']:
            new_task_def['containerDefinitions'][0]['memory'] = self.deploy_ctx['TASK_MEMORY']
        if 'TASK_SOFT_MEMORY' in self.deploy_ctx and self.deploy_ctx['TASK_SOFT_MEMORY']:
            new_task_def['containerDefinitions'][0]['memoryReservation'] = self.deploy_ctx['TASK_SOFT_MEMORY']
        if 'TASK_CPU' in self.deploy_ctx and self.deploy_ctx['TASK_CPU']:
            new_task_def['containerDefinitions'][0]['cpu'] = self.deploy_ctx['TASK_CPU']
        updated_task_definition = self.client.register_task_definition(**new_task_def)['taskDefinition']
        new_task_def_arn = updated_task_definition['taskDefinitionArn']
        self.deploy_ctx.notify_event(
            title="Update of ecs service {service} started".format(service=self.ecs_service),
            type="success")
        self.client.update_service(
            cluster=self.cluster,
            service=self.ecs_service,
            taskDefinition=new_task_def_arn)
        waiter = self.client.get_waiter('services_stable')
        success = True
        try:
            waiter.wait(cluster=self.cluster,
                        services=[
                            self.ecs_service
                        ]
                        )
        except Exception as e:
            success = False
            print_utility.error("Error waiting for service to stabilize - {}".format(e.message),raise_exception=True)
        finally:
            self.deploy_ctx.notify_event(
                title="Update of ecs service {service} completed".format(service=self.ecs_service,
                                                                         success="Success" if success else "Failed"),
                type="success" if success else "error")

    def _describe_task_definition(self, refresh=False):
        if self.task_definition_description and not refresh: return
        self.task_definition_description = self.client.describe_task_definition(taskDefinition=self.ecs_task_family)[
            'taskDefinition']



import json
from collections import defaultdict
from pprint import pformat

import boto3
import botocore
import pydash as pydash
from botocore.exceptions import WaiterError

from infra_buddy.utility import print_utility
from infra_buddy.utility.exception import NOOPException
from infra_buddy.utility.waitfor import waitfor


def _load_file_to_json(parameter_file):
    with open(parameter_file, 'r') as source:
        return json.load(source)


class CloudFormationBuddy(object):
    def __init__(self, deploy_ctx):
        super(CloudFormationBuddy, self).__init__()
        self.exports = {}
        self.resources = []
        self.deploy_ctx = deploy_ctx
        self.client = boto3.client('cloudformation', region_name=self.deploy_ctx.region)
        self.existing_change_set_id = None
        self.stack_id = None
        self.change_set_description = None
        self.stack_description = None
        self.stack_name = self.deploy_ctx.stack_name

    def does_stack_exist(self):
        return self._describe_stack()

    def _describe_stack(self):
        try:
            stacks = self.client.describe_stacks(StackName=self.stack_name)['Stacks']
            if len(stacks) >= 1:
                print_utility.info("Stack Description - {}".format(pformat(stacks)))
                self.stack_description = stacks[0]
                self.stack_id = self.stack_description['StackId']
                return True
        except (botocore.exceptions.ValidationError, botocore.exceptions.ClientError) as err:
            pass
        return False

    def get_stack_status(self, refresh=False):
        if not self.stack_description or refresh:  self._describe_stack()
        return self.stack_description['StackStatus'] if self.stack_description else "UNAVAILABLE"

    def delete_change_set(self):
        self._validate_changeset_operation_ready('delete_change_set')
        if self.get_change_set_execution_status(refresh=True) == 'EXECUTE_FAILED':
            print_utility.info(
                "Skipping Delete ChangeSet - ChangeSetID: {} Execution Status Failed".format(
                    self.existing_change_set_id))
            return
        response = self.client.delete_change_set(ChangeSetName=self.existing_change_set_id)
        print_utility.info(
            "Deleted ChangeSet - ChangeSetID: {} Response: {}".format(self.existing_change_set_id, response))

    def create_change_set(self, template_file_url, parameter_file):
        resp = self.client.create_change_set(
            StackName=self.stack_name,
            TemplateURL=template_file_url,
            Parameters=_load_file_to_json(parameter_file),
            Capabilities=[
                'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'
            ],
            ChangeSetName=self.deploy_ctx.change_set_name
        )
        self.existing_change_set_id = resp['Id']
        self.stack_id = resp['StackId']
        print_utility.info("Created ChangeSet:\nChangeSetID: {}\nStackID: {}\n{}".format(resp['Id'],
                                                                                        resp['StackId'],
                                                                                        pformat(resp,indent=1)))
        waiter = self.client.get_waiter('change_set_create_complete')
        try:
            waiter.wait(ChangeSetName=self.deploy_ctx.change_set_name, StackName=self.stack_id)
        except WaiterError as we:
            self.change_set_description = we.last_response
            noop = self._is_noop_changeset()
            print_utility.info("ChangeSet Failed to Create - {}".format(self.change_set_description['StatusReason']))
            if not noop:
                self.log_changeset_status()
                self._clean_change_set_and_exit()

    def _is_noop_changeset(self):
        return "The submitted information didn't contain changes." in self.change_set_description.get('StatusReason','')

    def get_change_set_status(self, refresh=False):
        self._validate_changeset_operation_ready('get_change_set_status')
        self.describe_change_set(refresh=refresh)
        return self.change_set_description['Status']

    def describe_change_set(self,refresh=False):
        self._validate_changeset_operation_ready('describe_change_set')
        if not self.change_set_description or refresh:
            self.change_set_description = self.client.describe_change_set(ChangeSetName=self.existing_change_set_id)

    def get_change_set_execution_status(self, refresh=False):
        self._validate_changeset_operation_ready('get_change_set_execution_status')
        self.describe_change_set(refresh=refresh)
        return self.change_set_description['ExecutionStatus']

    def execute_change_set(self):
        self._validate_changeset_operation_ready('execute changeset')
        action = 'update-stack'
        self._start_update_event(action)
        self.client.execute_change_set(ChangeSetName=self.existing_change_set_id)
        waiter = self.client.get_waiter('stack_update_complete')
        try:
            waiter.wait(StackName=self.stack_id)
            success = True
        except WaiterError as we:
            success = False
        self._finish_update_event(action, success)
        if not success:
            self._clean_change_set_and_exit(failed=True,failure_stage='execute')

    def _validate_changeset_operation_ready(self, operation):
        if not self.existing_change_set_id:
            raise Exception("Attempted to {} before create was called".format(operation))

    def should_execute_change_set(self):
        self._validate_changeset_operation_ready('should_execute_change_set')
        self.describe_change_set()
        if self._is_noop_changeset():
            return False
        changes_ = self.change_set_description['Changes']
        if len(changes_) == 2:
            if pydash.get(changes_[0], 'ResourceChange.ResourceType') == "AWS::ECS::Service":
                if pydash.get(changes_[1], 'ResourceChange.ResourceType') == "AWS::ECS::TaskDefinition":
                    if self.deploy_ctx.should_skip_ecs_trivial_update():
                        print_utility.info(
                            "WARN: Skipping changeset update because no computed changes except to service & task "
                            "rerun with SKIP_ECS=True to force")
                        return False
        return True

    def should_create_change_set(self):
        exists = self.does_stack_exist()
        if exists:
            if self.get_stack_status() == 'ROLLBACK_COMPLETE':
                print_utility.error("Can not update stack in state 'ROLLBACK_COMPLETE' -"
                                    " delete stack to recreate.",
                                    raise_exception=True)
        return exists

    def create_stack(self, template_file_url, parameter_file):
        action = 'create-stack'
        self._start_update_event(action)
        print_utility.info("Template URL: " + template_file_url)
        resp = self.client.create_stack(
            StackName=self.stack_name,
            TemplateURL=template_file_url,
            Parameters=_load_file_to_json(parameter_file),
            Capabilities=[
                'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'
            ],
            Tags=[
                {
                    'Key': 'Environment',
                    'Value': self.deploy_ctx.environment
                },
                {
                    'Key': 'Application',
                    'Value': self.deploy_ctx.application
                },
                {
                    'Key': 'Role',
                    'Value': self.deploy_ctx.role
                }
            ]
        )
        self.stack_id = resp['StackId']
        waiter = self.client.get_waiter('stack_create_complete')
        try:
            waiter.wait(StackName=self.stack_id)
            success = True
        except WaiterError as we:
            self.stack_description = we.last_response
            success = False
        self._finish_update_event(action, success)
        print_utility.info("Created Stack -  StackID: {}".format(resp['StackId']))
        if not success:
            raise Exception("Cloudformation stack failed to create")

    def log_stack_status(self, print_stack_events=False):
        print_utility.banner_warn("Stack Details: {}".format(self.stack_id), pformat(self.stack_description,indent=2))
        if print_stack_events:
            self._print_stack_events()

    def log_changeset_status(self, warn=True):
        if warn:
            print_utility.banner_warn("ChangeSet Details: {}".format(self.existing_change_set_id),
                                  pformat(self.change_set_description))
        else:
            print_utility.info("ChangeSet Details: {}".format(self.existing_change_set_id))
            print_utility.info_banner(pformat(self.change_set_description))

    def _clean_change_set_and_exit(self, failed=False, failure_stage='create'):
        self.delete_change_set()
        if failed:
            raise Exception("FAILED: Could not {} changeset".format(failure_stage))

    def _start_update_event(self, action):
        self.deploy_ctx.notify_event(
            title="{action} stack {stack_name} started".format(action=action, stack_name=self.deploy_ctx.stack_name),
            type="success")

    def _finish_update_event(self, action, success):
        msg = "{action} stack {stack_name} {status}".format(action=action,
                                                            stack_name=self.deploy_ctx.stack_name,
                                                            status='completed' if success else 'failed')
        print_utility.progress(msg)
        if success:
            self.deploy_ctx.notify_event(title=msg,
                                         type="success")
        else:
            self.deploy_ctx.notify_event(title=msg,
                                         type="error")
            self.log_stack_status(print_stack_events=True)

    def get_export_value(self, param=None, fully_qualified_param_name=None):
        if not fully_qualified_param_name:
            fully_qualified_param_name = "{stack_name}-{param}".format(stack_name=self.stack_name, param=param)
        if len(self.exports) == 0: self._load_export_values()
        val = self.exports.get(fully_qualified_param_name, None)
        if val is None:
            print_utility.warn("Could not locate export value - {}".format(fully_qualified_param_name))
        return val

    def _load_export_values(self):
        export_results = self.client.list_exports()
        export_list = export_results['Exports']
        while export_list is not None:
            for export in export_list:
                self.exports[export['Name']] = export['Value']
            next_ = export_results.get('NextToken', None)
            if next_:
                export_results = self.client.list_exports(NextToken=next_)
                export_list = export_results.get('Exports', None)
            else:
                export_list = None

    def get_existing_parameter_value(self, param_val):
        self._describe_stack()
        for param in self.stack_description.get('Parameters', []):
            if param['ParameterKey'] == param_val:
                return param['ParameterValue']
        print_utility.error("Could not locate parameter value: {}".format(param_val))
        return None

    def get_resource_list(self):
        if len(self.resources) == 0:
            self.resources = self.load_stack_resources(self.stack_name)
        return self.resources

    def load_stack_resources(self, to_inspect):
        ret = []
        res = self.client.list_stack_resources(StackName=to_inspect)
        res_resources_list = res['StackResourceSummaries']
        while res_resources_list is not None:
            ret.extend(res_resources_list)
            next_ = res.get('NextToken', None)
            if next_:
                res = self.client.list_stack_resources(StackName=to_inspect, NextToken=next_)
                res_resources_list = res['StackResourceSummaries']
            else:
                res_resources_list = None
        return ret

    def list_stacks(self, filter=None):
        ret = []
        res = self.client.list_stacks(StackStatusFilter=['UPDATE_COMPLETE', 'CREATE_COMPLETE'])
        res_stack_list = res['StackSummaries']
        while res_stack_list is not None:
            ret.extend(res_stack_list)
            next_ = res.get('NextToken', None)
            if next_:
                res = self.client.list_stacks(StackStatusFilter=['UPDATE_COMPLETE', 'CREATE_COMPLETE'], NextToken=next_)
                res_stack_list = res['StackSummaries']
            else:
                res_stack_list = None
        pluck = pydash.pluck(ret, "StackName")
        return [stack for stack in pluck if filter and stack.startswith(filter)]

    def load_resources_for_stack_list(self, stacks):
        resources = defaultdict(list)
        for stack in stacks:
            resources[stack] = self.load_stack_resources(stack)
        return resources

    def _print_stack_events(self):
        events = []
        res = self.client.describe_stack_events(StackName=self.stack_name)
        res_list = res['StackEvents']
        while res_list is not None:
            events.extend(res_list)
            next_ = res.get('NextToken', None)
            if next_:
                res = self.client.describe_stack_events(StackName=self.stack_name, NextToken=next_)
                res_list = res['StackEvents']
            else:
                res_list = None
        for ev in events:
            if "ResourceStatusReason" in ev:
                template = "{Timestamp}\t{ResourceStatus}\t{ResourceType}\t{LogicalResourceId}\t{ResourceStatusReason}"
            else:
                template = "{Timestamp}\t{ResourceStatus}\t{ResourceType}\t{LogicalResourceId}"
            print_utility.warn(template.format(**ev))



import click

from infra_buddy.commandline import cli
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.deploy.cloudformation_deploy import CloudFormationDeploy
from infra_buddy.template.template import NamedLocalTemplate
from infra_buddy.utility import print_utility


@cli.command(name='validate-template',short_help="Analyze a service template or built-in service-type for logical "
                                                 "errors and usability issues.")
@click.option("--service-template-directory", type=click.Path(exists=True), help="The directory containing "
                                                                         "the service template.")
@click.option("--service-type", help="The service-type of an internal template.")
@click.pass_obj
def deploy_cloudformation(deploy_ctx, service_template_directory, service_type):
    # type: (DeployContext,str,str) -> None
    do_command(deploy_ctx, service_template_directory, service_type)


def do_command(deploy_ctx, service_template_directory=None, service_type=None):
    # type: (DeployContext,[str or None],str) -> None
    if service_template_directory is None:
        print_utility.warn(
            "Service template directory was not provided.  Assuming service-type '{}' is built-in.".format(
                service_type))
        template = deploy_ctx.template_manager.get_known_template(template_name=service_type)
        deploy = CloudFormationDeploy(stack_name=deploy_ctx.stack_name, template=template, deploy_ctx=deploy_ctx)
    else:
        deploy = CloudFormationDeploy(stack_name=deploy_ctx.stack_name,
                                      template=NamedLocalTemplate(service_template_directory ),
                                      deploy_ctx=deploy_ctx)
    errs = deploy.analyze()
    if errs > 0:
        print_utility.error("Template raised {} errors ".format(errs), raise_exception=True)



import json
import os

from jsonschema import validate

from infra_buddy.deploy.ecs_deploy import ECSDeploy
from infra_buddy.deploy.s3_deploy import S3Deploy
from infra_buddy.utility import print_utility

_ARTIFACT_FILE = "artifact.json"

_CONTAINER_ARTIFACT_TYPE = "container"
_S3_ARTIFACT_TYPE = "s3"
_NOOP_ARTIFACT_TYPE = "noop"

_ARTIFACT_TYPE = "artifact-type"
_ARTIFACT_IDENTIFIER = "artifact-identifier"
_ARTIFACT_LOCATION = "artifact-path"


class ArtifactDefinition(object):
    schema = {
        "type": "object",
        "properties": {
            _ARTIFACT_TYPE: {"type": "string", "enum": [_CONTAINER_ARTIFACT_TYPE, _S3_ARTIFACT_TYPE]},
            _ARTIFACT_IDENTIFIER: {"type": "string"},
            _ARTIFACT_LOCATION: {"type": "string"}
        },
        "required": [
            _ARTIFACT_TYPE,
            _ARTIFACT_IDENTIFIER,
            _ARTIFACT_LOCATION
        ]
    }

    @classmethod
    def create(cls, artifact_type=None, artifact_location=None, artifact_identifier=None):
        # type: (str, str, str) -> ArtifactDefinition
        if artifact_type == _CONTAINER_ARTIFACT_TYPE:
            return ECSArtifactDefinition(artifact_location=artifact_location,
                                         artifact_identifier=artifact_identifier)
        elif artifact_type == _S3_ARTIFACT_TYPE:
            return S3ArtifactDefinition(artifact_location=artifact_location,
                                        artifact_identifier=artifact_identifier)
        elif artifact_type:
            raise Exception("Unknown artifact type")
        else:
            return NOOPArtifactDefinition(None, None)

    @classmethod
    def create_from_directory(cls, artifact_directory):
        # type: (str) -> ArtifactDefinition
        definition = ArtifactDefinition._load_artifact_definition(artifact_directory)
        if not definition:
            definition = ArtifactDefinition._search_for_legacy_implementation(artifact_directory)
        if definition:
            validate(definition, ArtifactDefinition.schema)
            return cls.create(definition[_ARTIFACT_TYPE], definition[_ARTIFACT_LOCATION],
                              definition[_ARTIFACT_IDENTIFIER])
        else:
            return cls.create()

    def __init__(self, artifact_type, artifact_location, artifact_identifier):
        super(ArtifactDefinition, self).__init__()
        self.artifact_type = artifact_type
        self.artifact_location = artifact_location
        self.artifact_id = artifact_identifier

    @staticmethod
    def _search_for_legacy_implementation(artifact_directory):
        image_definition = os.path.join(artifact_directory, "containerurl.txt")
        ret = {_ARTIFACT_TYPE: _CONTAINER_ARTIFACT_TYPE}
        if os.path.exists(image_definition):
            print_utility.warn("Defining ECS service udpate with deprecated containerurl.txt artifact")
            with open(image_definition, 'r') as image:
                image_def = image.read().strip()
                rfind = image_def.rfind(":")
                ret[_ARTIFACT_LOCATION] = image_def[:rfind]
                ret[_ARTIFACT_IDENTIFIER] = image_def[rfind + 1:]
            return ret
        return None

    @staticmethod
    def _load_artifact_definition(artifact_directory):
        artifact_def_path = os.path.join(artifact_directory, _ARTIFACT_FILE)
        if os.path.exists(artifact_def_path):
            print_utility.info("Defining artifact definition with artifact.json - {}".format(artifact_def_path))
            with open(artifact_def_path, 'r') as art_def:
                return json.load(art_def)
        else:
            print_utility.warn("Artifact definition (artifact.json) did not exist in artifact directory."
                               " Continuing infrastructure update without artifact deploy.")
        return None

    def generate_execution_plan(self, deploy_ctx):
        raise Exception("Well, this is unexpected")

    def save_to_file(self, destination_dir=None):
        if destination_dir:
            path = os.path.join(destination_dir, _ARTIFACT_FILE)
        else:
            path = _ARTIFACT_FILE
        print_utility.info("Persisting artifact manfiest - {}".format(path))
        with open(path, 'w') as file:
            json.dump({_ARTIFACT_TYPE: self.artifact_type,
                       _ARTIFACT_LOCATION: self.artifact_location,
                       _ARTIFACT_IDENTIFIER: self.artifact_id}, file)
        return path

    def register_env_variables(self, deploy_ctx):
        if self.artifact_type == _CONTAINER_ARTIFACT_TYPE:
            # For first time deploys of ECS services
            deploy_ctx['IMAGE'] = "{location}:{tag}".format(location=self.artifact_location, tag=self.artifact_id)


class ECSArtifactDefinition(ArtifactDefinition):
    def __init__(self, artifact_location, artifact_identifier):
        super(ECSArtifactDefinition, self).__init__(_CONTAINER_ARTIFACT_TYPE, artifact_location, artifact_identifier)

    def generate_execution_plan(self, deploy_ctx):
        return [ECSDeploy(self.artifact_id, self.artifact_location, deploy_ctx)]


class S3ArtifactDefinition(ArtifactDefinition):
    def __init__(self, artifact_location, artifact_identifier):
        super(S3ArtifactDefinition, self).__init__(_S3_ARTIFACT_TYPE, artifact_location, artifact_identifier)

    def generate_execution_plan(self, deploy_ctx):
        return [S3Deploy(self.artifact_id, self.artifact_location, deploy_ctx)]


class NOOPArtifactDefinition(ArtifactDefinition):
    def __init__(self, artifact_location, artifact_identifier):
        super(NOOPArtifactDefinition, self).__init__(_NOOP_ARTIFACT_TYPE, artifact_location, artifact_identifier)

    def generate_execution_plan(self, deploy_ctx):
        return None



import os
import tempfile
from zipfile import ZipFile

import requests
from copy import deepcopy

from infra_buddy.aws import s3
from infra_buddy.context import monitor_definition
from infra_buddy.context.monitor_definition import MonitorDefinition
from infra_buddy.utility import print_utility


class Template(object):
    def __init__(self, service_type, values, template_name='cloudformation'):
        super(Template, self).__init__()
        self.template_name = template_name
        self.service_type = service_type
        self.destination_relative = None
        self.destination = None
        self.valid = False
        self.default_env_values = values.get('default-values',{})

    def get_default_env_values(self):
        return self.default_env_values


    def get_parameter_file_path(self):
        return os.path.join(self._get_template_location(),
                            "{template_name}.parameters.json".format(template_name=self.template_name))

    def get_defaults_file_path(self):
        return os.path.join(self._get_template_location(),
                            "defaults.json".format(service_type=self.service_type))

    def get_template_file_path(self):
        return os.path.join(self._get_template_location(),
                            "{template_name}.template".format(template_name=self.template_name))

    def get_config_dir(self):
        config_path = os.path.join(self._get_template_location(), "config")
        return config_path if os.path.exists(config_path) else None

    def _get_template_location(self):
        return os.path.join(self.destination,
                            self.destination_relative) if self.destination_relative else self.destination

    def _validate_template_dir(self, err_on_failure_to_locate=True):
        if not os.path.exists(self.get_template_file_path()):
            if err_on_failure_to_locate: print_utility.error("Template file could not be "
                                                             "located for service - {service_type}".format(
                service_type=self.service_type), raise_exception=True)
            return
        if not os.path.exists(self.get_parameter_file_path()):
            if err_on_failure_to_locate: print_utility.error("Parameter file could not be "
                                                             "located for service - {service_type}".format(
                service_type=self.service_type), raise_exception=True)
            return
        self.valid = True

    def _prep_download(self):
        if not self.destination:
            self.destination = tempfile.mkdtemp()

    def _set_download_relative_path(self, path):
        self.destination_relative = path

    def has_monitor_definition(self):
        return os.path.exists(self.get_monitor_definition_file())

    def get_monitor_artifact(self):
        return MonitorDefinition.create_from_directory(self._get_template_location())

    def get_monitor_definition_file(self):
        return os.path.join(self._get_template_location(),monitor_definition._ARTIFACT_FILE)


class URLTemplate(Template):
    def __init__(self, service_type, values):
        super(URLTemplate, self).__init__(service_type, values)
        self.download_url = values.get('url', None)

    def download_template(self):
        self._prep_download()
        r = requests.get(self.download_url, stream=True)
        temporary_file = tempfile.NamedTemporaryFile()
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                temporary_file.write(chunk)
        temporary_file.seek(0)
        with ZipFile(temporary_file) as zf:
            zf.extractall(self.destination)
        self._validate_template_dir()


class AliasTemplate(Template):

    def __init__(self, service_type, values):
        super(AliasTemplate, self).__init__(service_type, values=values)
        self.lookup = values.get('lookup', None)
        self.delegate = None

    def resolve(self,templates):
        # type: (dict()) -> None
        if self.lookup in templates:
            self.delegate = templates[self.lookup]
        else:
            raise Exception("Unable to resolve alias template - {}".format(self.lookup))

    def download_template(self):
        self.delegate.download_template()

    def has_monitor_definition(self):
        return self.delegate.has_monitor_definition()

    def get_monitor_artifact(self):
        return self.delegate.get_monitor_artifact()

    def get_monitor_definition_file(self):
        return self.delegate.get_monitor_definition_file()

    def get_template_file_path(self):
        return self.delegate.get_template_file_path()

    def get_defaults_file_path(self):
        return self.delegate.get_defaults_file_path()

    def get_config_dir(self):
        return self.delegate.get_config_dir()

    def get_parameter_file_path(self):
        return self.delegate.get_parameter_file_path()

    def get_default_env_values(self):
        values = deepcopy(self.delegate.get_default_env_values())
        values.update(self.default_env_values)
        return values


class GitHubTemplate(URLTemplate):
    def __init__(self, service_type, values):
        super(GitHubTemplate, self).__init__(service_type=service_type, values=values)
        tag = values.pop('tag', 'master')
        self.download_url = "https://github.com/{owner}/{repo}/archive/{tag}.zip".format(tag=tag, **values)
        self._set_download_relative_path("{repo}-{tag}".format(tag=tag, **values))


class NamedLocalTemplate(Template):
    def __init__(self, directory, service_type="local-template", err_on_failure_to_locate=True, template_name="cloudformation"):
        super(NamedLocalTemplate, self).__init__(service_type, values={},template_name=template_name)
        self.destination = directory
        self._validate_template_dir(err_on_failure_to_locate=err_on_failure_to_locate)


class S3Template(Template):
    def __init__(self, service_type, values):
        super(S3Template, self).__init__(service_type, values=values)
        self.s3_location = values['location']

    def download_template(self):
        self._prep_download()
        s3.download_zip_from_s3_url(self.s3_location, self.destination)


class LocalTemplate(Template):
    def __init__(self, template, parameter_file, config_dir=None):
        super(LocalTemplate, self).__init__("", values={})
        self.config_dir = config_dir
        self.parameter_file = parameter_file
        self.template = template

    def get_parameter_file_path(self):
        return self.parameter_file

    def get_defaults_file_path(self):
        return None

    def get_template_file_path(self):
        return self.template

    def get_config_dir(self):
        return self.config_dir



import os
from collections import defaultdict

import boto3
import click
from infra_buddy.aws.cloudformation import CloudFormationBuddy

from infra_buddy.commandline import cli
from infra_buddy.context.deploy_ctx import DeployContext
from infra_buddy.deploy.cloudformation_deploy import CloudFormationDeploy
from infra_buddy.template.template import NamedLocalTemplate
from infra_buddy.utility import print_utility


@cli.command(name='introspect', short_help="Search infra-buddy managed services for a resource.")
@click.option("--type-filter", help="Constrain search to a AWS resource type.")
@click.pass_obj
def deploy_cloudformation(deploy_ctx,type_filter):
    # type: (DeployContext,str) -> None
    do_command(deploy_ctx,type_filter)


def do_command(deploy_ctx,type_filter):
    # type: (DeployContext,str) -> None
    cf_buddy = CloudFormationBuddy(deploy_ctx=deploy_ctx)
    stacks = cf_buddy.list_stacks(deploy_ctx.stack_name)
    resources = cf_buddy.load_resources_for_stack_list(stacks)
    for stack_name, resources in resources.iteritems():
        print_utility.banner("Stack: {}".format(stack_name))
        for resource in resources:
            if not type_filter or type_filter in resource['ResourceType']:
                print_utility.info_banner("\tName: {}".format(resource['LogicalResourceId']))
                print_utility.info_banner("\tType: {}".format(resource['ResourceType']))

        
        




